/********************************************************************
* Description: cincikins.c
*   Kinematics for Stuart Stevenson's 5 axis mill hydrotel retrofit.
*   This mill has an AB tilting head with A axis riding on B.
*   And B riding on Z.
*
* Author: Stuart Stevenson
* License: GPL Version 2
*    
* Copyright (c) 2007 Stuart Stevenson
********************************************************************/

#include "kinematics.h"		/* these decls */
#include "posemath.h"
#include "hal.h"
#include "rtapi.h"
#include "rtapi_math.h"

#define d2r(d) ((d)*PM_PI/180.0)
#define r2d(r) ((r)*180.0/PM_PI)

#ifndef hypot
#define hypot(a,b) (sqrt((a)*(a)+(b)*(b)))
#endif

struct haldata {
    hal_float_t pivot_length;
//    hal_float_t tool_length;
    hal_float_t aerry;
    hal_float_t berrx;
    hal_float_t berrz;
    hal_float_t yskewx;
    hal_float_t yskewz;
    hal_float_t zskewx;
    hal_float_t zskewy;
    hal_float_t askew;
    hal_float_t bskew;

} *haldata;

int kinematicsForward(const double *joints,
		      EmcPose * pos,
		      const KINEMATICS_FORWARD_FLAGS * fflags,
		      KINEMATICS_INVERSE_FLAGS * iflags)
{
    // AB angle correction
            double joint3radians = d2r(joints[3]);
            double joint4radians = d2r(joints[4]);

            double askewAngle = atan2(haldata->askew,1.0);
            double bskewAngleNeg = atan2(-haldata->bskew,1.0);
            double bskewAngle = atan2(haldata->bskew,1.0);
            double sinAskewAngle = sin(askewAngle);
            double cosAskewAngle = cos(askewAngle);
            double cosBskewAngleNeg = cos(bskewAngleNeg);
            double sinJoint3 = sin(joint3radians);
            double cosJoint3 = cos(joint3radians);
            double sinJoint4 = sin(joint4radians);
            double cosJoint4 = cos(joint4radians);
            double fva11 = sinJoint3 * sinAskewAngle;
            double fva12 = sinJoint3 * cosAskewAngle;
            double fva14 = fva11 / cosJoint4;
            double fva15 = fva14 * sinJoint4;
            double fva16 = fva15 - cosJoint3;
            double fva17 = fva16 * cosJoint4;
            double fva18 = fva16 * sinJoint4;
            double fva19 = fva12 * -haldata->bskew;
            double fva20 = fva18 - fva19 - fva14;
            double fva22 = fva20 * cosBskewAngleNeg;
            double fbcomp = r2d(atan2(-fva22,-fva17));
            double fva25 = fva12 / cosBskewAngleNeg;
            double fva26 = fva22 * -haldata->bskew;
            double fva27 = fva25 + fva26;
            double facomp = r2d(atan2(fva27,sqrt(1-fva27*fva27)));

    // XYZ geometry correction
            double sinBskewAngle = sin(bskewAngle);
            double cosBskewAngle = cos(bskewAngle);
            double fva39 = haldata->berrz * sinJoint4;
            double fva40 = haldata->berrx * cosJoint4;
            double fva41 = fva39 + fva40;
            double fva42 = fva41 * cosBskewAngle;
            double fva43 = haldata->aerry * cosJoint3;
            double fva44 = fva43 * cosAskewAngle;
            double fva45 = fva44 * sinBskewAngle;
            double fva46 = fva43 * sinAskewAngle;
            double fva47 = fva46 * cosJoint4;
            double fva48 = fva47 * -cosBskewAngle;
            double fva49 = haldata->aerry * sinJoint3;
            double fva50 = fva49 * sinJoint4;
            double fva51 = fva50 * cosBskewAngle;
            double fva52 = fva41 * -sinBskewAngle;
            double fva53 = fva44 * cosBskewAngle;
            double fva54 = fva47 * sinBskewAngle;
            double fva55 = fva50 * -sinBskewAngle;
            double fva56 = haldata->berrz * cosJoint4;
            double fva57 = haldata->berrx * -sinJoint4;
            double fva58 = fva46 * sinJoint4;
            double fva59 = fva49 * cosJoint4;

            double fxcomp = fva42 + fva45 + fva48 + fva51;
            double fycomp = fva52 + fva53 + fva54 + fva55;
            double fzcomp = fva56 + fva57 + fva58 + fva59;

    // pivot length correction
            double forwardpivotlengthcompX = (haldata->pivot_length + joints[8]) * cos(d2r(pos->a)) * sin(d2r(pos->b));
            double forwardpivotlengthcompY = (haldata->pivot_length + joints[8]) * sin(d2r(pos->a));
            double forwardpivotlengthcompZ = (haldata->pivot_length + joints[8]) * cos(d2r(pos->a)) * cos(d2r(pos->b)) + haldata->pivot_length;

     // YZ skew correction - starting assumption of X axis/X joint perfection
// Y motion 
            double yskewxAngle = atan2(haldata->yskewx,1);
            double hypotyskewxAngle = pos->tran.y / cos(yskewxAngle);
//          Y compensation for XY orthogonal error on the XY plane
            double forwardYskewcompXY = hypotyskewxAngle - pos->tran.y;
//          X compensation for XY orthogonal error on the XY plane
            double forwardYskewcompYX = hypotyskewxAngle * sin(yskewxAngle);

// Z motion
            double zskewxAngle = atan2(haldata->zskewx,1);
            double hypotzskewxAngle = (pos->tran.z - haldata->pivot_length) / cos(zskewxAngle);
//          Z compensation for ZX orthogonal error on the ZX plane
            double forwardZskewcompXZ = hypotzskewxAngle - pos->tran.z;
//          X compensation for ZX orthogonal error on the ZX plane
            double forwardZskewcompZX = hypotzskewxAngle * sin(zskewxAngle);

// Z motion
            double zskewyAngle = atan2(haldata->zskewy,1);
            double hypotzskewyAngle = (pos->tran.z - haldata->pivot_length) / cos(zskewyAngle);
//          Z compensation for YZ orthogonal error on the YZ plane
            double forwardZskewcompYZ = hypotzskewyAngle - pos->tran.z;
//          Y compensation for YZ orthogonal error on the YZ plane
            double forwardZskewcompZY = hypotzskewyAngle * sin(zskewyAngle);

// Y motion
            double yskewzAngle = atan2(haldata->yskewz,1);
            double hypotyskewzAngle = pos->tran.y / cos(yskewzAngle);
//          Y compensation for YZ orthogonal error on the YZ plane
            double forwardYskewcompZY = hypotyskewzAngle - pos->tran.y;
//          Z compensation for YZ orthogonal error on the YZ plane
            double forwardYskewcompYZ = hypotyskewzAngle * sin(yskewzAngle);

   pos->tran.x = joints[0] - fxcomp - forwardpivotlengthcompX - forwardYskewcompYX - forwardZskewcompZX;
    pos->tran.y = joints[1] - fycomp + forwardpivotlengthcompY - forwardYskewcompXY - forwardZskewcompZY - forwardYskewcompZY;
    pos->tran.z = joints[2] - fzcomp - forwardpivotlengthcompZ - forwardZskewcompXZ - forwardZskewcompYZ - forwardYskewcompYZ;
    pos->a = facomp;
    pos->b = fbcomp;
    pos->c = joints[5];
    pos->u = joints[6];
    pos->v = joints[7];
    pos->w = joints[8];

    return 0;
}

int kinematicsInverse(const EmcPose * pos,
		      double *joints,
		      const KINEMATICS_INVERSE_FLAGS * iflags,
		      KINEMATICS_FORWARD_FLAGS * fflags)
{
    // AB angle correction
            double posaRadians = d2r(pos->a);
            double posbRadians = d2r(pos->b);

            double askewAngle = atan2(haldata->askew,1);
            double bskewAngleNeg = atan2(-haldata->bskew,1);
            double bskewAngle = atan2(haldata->bskew,1);
            double sinPosA = sin(posaRadians);
            double cosPosA = cos(posaRadians);
            double iva11 = cosPosA * sin(-posbRadians);
            double iva12 = iva11 / cos(bskewAngleNeg);
            double iva13 = iva12 * sin(bskewAngleNeg);
            double iva14 = sinPosA - iva13;
            double iva15 = iva14 * cos(bskewAngleNeg);
            double iva16 = iva15 / cos(askewAngle);
            double iva17 = atan2(iva16,sqrt((1 - iva16 * iva16)));
            double iva18 = cos(iva17);
            double iacomp = r2d(iva17);
            double iva19 = iva18 * sin(posbRadians);
            double iva20 = iva16 * sin(askewAngle);
            double iva21 = iva20 * cos(posbRadians);
            double iva22 = -iva19 - iva21;
            double iva23 = r2d(atan2(iva15,sqrt((1 - iva15 * iva15))));
            double iva24 = cos(d2r(iva23));
            double iva25a = iva22 / iva24;
            double iva25 = r2d(atan2(iva25a,sqrt(1 - iva25a * iva25a)));
            double iva26 = iva15 * -haldata->bskew;
            double iva27 = iva26 + iva12;
            double iva28a = iva27 / iva24;
            double iva28 = r2d(atan2(iva28a,sqrt((1 - iva28a * iva28a))));
            double ibcomp = r2d(posbRadians) - iva28 + iva25;

    // XYZ geometry correction
            double sinCompensatedAngleA = sin(d2r(iacomp));
            double cosCompensatedAngleA = cos(d2r(iacomp));
            double sinCompensatedAngleB = sin(d2r(ibcomp));
            double cosCompensatedAngleB = cos(d2r(ibcomp));
            double sinAskewAngle = sin(askewAngle);
            double cosAskewAngle = cos(askewAngle);
            double sinBskewAngle = sin(bskewAngle);
            double cosBskewAngle = cos(bskewAngle);
            double iva39 = haldata->berrz * sinCompensatedAngleB;
            double iva40 = haldata->berrx * cosCompensatedAngleB;
            double iva41 = iva39 + iva40;
            double iva42 = iva41 * cosBskewAngle;
            double iva43 = haldata->aerry * cosCompensatedAngleA;
            double iva44 = iva43 * cosAskewAngle;
            double iva45 = iva44 * sinBskewAngle;
            double iva46 = iva43 * sinAskewAngle;
            double iva47 = iva46 * cosCompensatedAngleB;
            double iva48 = iva47 * -cosBskewAngle;
            double iva49 = haldata->aerry * sinCompensatedAngleA;
            double iva50 = iva49 * sinCompensatedAngleB;
            double iva51 = iva50 * cosBskewAngle;
            double iva52 = iva41 * -sinBskewAngle;
            double iva53 = iva44 * cosBskewAngle;
            double iva54 = iva47 * sinBskewAngle;
            double iva55 = iva50 * -sinBskewAngle;
            double iva56 = haldata->berrz * cosCompensatedAngleB;
            double iva57 = haldata->berrx * -sinCompensatedAngleB;
            double iva58 = iva46 * sinCompensatedAngleB;
            double iva59 = iva49 * cosCompensatedAngleB;

            double ixcomp = iva42 + iva45 + iva48 + iva51;
            double iycomp = iva52 + iva53 + iva54 + iva55;
            double izcomp = iva56 + iva57 + iva58 + iva59;

    // pivot length correction
            double inversepivotlengthcompX = ((haldata->pivot_length + pos->w) * cos(d2r(pos->a)) * sin(d2r(pos->b)));
            double inversepivotlengthcompY = ((haldata->pivot_length + pos->w) * sin(d2r(pos->a)));
            double inversepivotlengthcompZ = ((haldata->pivot_length + pos->w) * cos(d2r(pos->a)) * cos(d2r(pos->b))) + haldata->pivot_length;

    // YZ skew correction - starting assumption of X axis/X joint perfection
// Y motion 
            double yskewxAngle = atan2(haldata->yskewx,1);
            double hypotyskewxAngle = pos->tran.y / cos(yskewxAngle);
//          Y compensation for XY orthogonal error on the XY plane
            double inverseYskewcompXY = hypotyskewxAngle - pos->tran.y;
//          X compensation for XY orthogonal error on the XY plane
            double inverseYskewcompYX = hypotyskewxAngle * sin(yskewxAngle);

// Z motion
            double zskewxAngle = atan2(haldata->zskewx,1);
            double hypotzskewxAngle = (pos->tran.z - haldata->pivot_length) / cos(zskewxAngle);
//          Z compensation for ZX orthogonal error on the ZX plane
            double inverseZskewcompXZ = hypotzskewxAngle - pos->tran.z;
//          X compensation for ZX orthogonal error on the ZX plane
            double inverseZskewcompZX = hypotzskewxAngle * sin(zskewxAngle);

// Z motion
            double zskewyAngle = atan2(haldata->zskewy,1);
            double hypotzskewyAngle = (pos->tran.z - haldata->pivot_length) / cos(zskewyAngle);
//          Z compensation for YZ orthogonal error on the YZ plane
            double inverseZskewcompYZ = hypotzskewyAngle - pos->tran.z;
//          Y compensation for YZ orthogonal error on the YZ plane
            double inverseZskewcompZY = hypotzskewyAngle * sin(zskewyAngle);

// Y motion
            double yskewzAngle = atan2(haldata->yskewz,1);
            double hypotyskewzAngle = pos->tran.y / cos(yskewzAngle);
//          Y compensation for YZ orthogonal error on the YZ plane
            double inverseYskewcompZY = hypotyskewzAngle - pos->tran.y;
//          Z compensation for YZ orthogonal error on the YZ plane
            double inverseYskewcompYZ = hypotyskewzAngle * sin(yskewzAngle);

    joints[0] = pos->tran.x + ixcomp + inversepivotlengthcompX + inverseYskewcompYX + inverseZskewcompZX;
    joints[1] = pos->tran.y + iycomp - inversepivotlengthcompY + inverseYskewcompXY + inverseZskewcompZY + inverseYskewcompZY;
    joints[2] = pos->tran.z + izcomp + inversepivotlengthcompZ + inverseZskewcompXZ + inverseZskewcompYZ + inverseYskewcompYZ;
    joints[3] = iacomp;
    joints[4] = ibcomp;
    joints[5] = pos->c;
    joints[6] = pos->u;
    joints[7] = pos->v;
    joints[8] = pos->w;

    return 0;
}

KINEMATICS_TYPE kinematicsType()
{
    return KINEMATICS_BOTH;
}

#ifdef RTAPI
#include "rtapi.h"		/* RTAPI realtime OS API */
#include "rtapi_app.h"		/* RTAPI realtime module decls */

EXPORT_SYMBOL(kinematicsType);
EXPORT_SYMBOL(kinematicsInverse);
EXPORT_SYMBOL(kinematicsForward);
MODULE_LICENSE("GPL");

int comp_id;
int rtapi_app_main(void) {
    int result;
    comp_id = hal_init("cincikins");
    if(comp_id < 0) return comp_id;

    haldata = hal_malloc(sizeof(struct haldata));

    result = hal_param_float_new("cincikins.pivot-length", \
             HAL_RW, &(haldata->pivot_length), comp_id);
    if(result < 0) goto error;
//    result = hal_param_float_new("cincikins.tool-length", \
//             HAL_RW, &(haldata->tool_length), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.aerry", \
             HAL_RW, &(haldata->aerry), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.berrx", \
             HAL_RW, &(haldata->berrx), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.berrz", \
             HAL_RW, &(haldata->berrz), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.yskewx", \
             HAL_RW, &(haldata->yskewx), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.yskewz", \
             HAL_RW, &(haldata->yskewz), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.zskewx", \
             HAL_RW, &(haldata->zskewx), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.zskewy", \
             HAL_RW, &(haldata->zskewy), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.askew", \
             HAL_RW, &(haldata->askew), comp_id);
    if(result < 0) goto error;
    result = hal_param_float_new("cincikins.bskew", \
             HAL_RW, &(haldata->bskew), comp_id);
    if(result < 0) goto error;

    haldata->pivot_length = 1.0;
//    haldata->tool_length = 1.0;
    haldata->aerry = 0;
    haldata->berrx = 0;
    haldata->berrz = 0;
    haldata->yskewx = 0;
    haldata->yskewz = 0;
    haldata->zskewx = 0;
    haldata->zskewy = 0;
    haldata->askew = 0;
    haldata->bskew = 0;

    hal_ready(comp_id);
    return 0;

error:
    hal_exit(comp_id);
    return result;
}

void rtapi_app_exit(void) { hal_exit(comp_id); }
#endif
